/*
 * Support for executing code in new tasks and threads.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>
#include <l4/sys/factory.h>
#include <l4/sys/task.h>
#include <l4/sys/thread.h>
#include <l4/util/bitops.h> /* l4util_log2 */

#include <fsclient/client.h>
#include <ipc/cap_alloc.h>
#include <ipc/mem_ipc.h>
#include <ipc/server.h>
#include <systypes/fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread-l4.h>
#include <pthread.h>

#include "system_pager_interface.h"
#include "system_pager_server.h"



/* A simple system pager. */

class ExecPager : public SystemPager
{
public:
  l4_addr_t buf, buf_start;
  unsigned int buf_log2size;

  virtual long exception(l4_exc_regs_t regs,
                         l4_snd_fpage_t *region);

  virtual long page_fault(l4_umword_t pfa, l4_umword_t pc,
                          l4_snd_fpage_t *region);
};

long ExecPager::exception(l4_exc_regs_t regs, l4_snd_fpage_t *region)
{
  (void) region;

  printf("exception(...) -> pfa = %lx, pc = %lx\n", l4_utcb_exc_pfa(&regs), l4_utcb_exc_pc(&regs));
  return L4_EOK;
}

long ExecPager::page_fault(l4_umword_t pfa, l4_umword_t pc, l4_snd_fpage_t *region)
{
  l4_umword_t addr = pfa & ~7UL, flags = pfa & 7;

  printf("page_fault(%lx, %lx) -> %lx (%lx)...\n", pfa, pc, addr, flags);

  if ((addr >= buf_start) && (addr < buf_start + (1UL << buf_log2size)))
  {
    region->fpage = l4_fpage(buf, buf_log2size, L4_FPAGE_RX);
    region->snd_base = buf_start;

    return L4_EOK;
  }

  return -L4_ENOMEM;
}



static ExecPager exec_pager;

static void init_pager(ipc_server_config_type *config, l4_addr_t buf,
                       unsigned int buf_log2size, l4_addr_t buf_start)
{
  exec_pager.buf = buf;
  exec_pager.buf_log2size = buf_log2size;
  exec_pager.buf_start = buf_start;

  ipc_server_init_config(config);

  config->expected_items = SystemPager_expected_items;
  config->handler = (ipc_server_handler_type) handle_SystemPager;
  config->handler_obj = static_cast<SystemPager *>(&exec_pager);
}

static long start_pager(ipc_server_config_type *config, pthread_t thread)
{
  config->config_thread = 1;
  config->thread = pthread_l4_cap(thread);

  printf("Starting pager thread...\n");
  return ipc_server_start_config(config);
}



/* UTCB properties.
   See: moe/server/src/loader.cc */

enum ipc_exec_utcb
{
  Default_max_threads = 16,
#ifdef ARCH_mips
  Utcb_area_start = 0x73000000,
#else
  Utcb_area_start = 0xb3000000,
#endif
};



static int find_region_exp(offset_t size)
{
  int exp = l4util_log2(size);

  if ((1UL << exp) < size)
    return exp + 1;
  else
    return exp;
}

static offset_t find_region_size(offset_t size)
{
  return 1 << find_region_exp(size);
}

int main(int argc, char *argv[])
{
  long err;

  if (argc < 2)
  {
    printf("Need a program to run.\n");
    return 1;
  }

  /* Allocate capabilities for the task and thread. */

  l4_cap_idx_t task, thread;

  task = ipc_cap_alloc();

  if (l4_is_invalid_cap(task))
  {
    printf("Could not allocate task.\n");
    return 1;
  }

  thread = ipc_cap_alloc();

  if (l4_is_invalid_cap(thread))
  {
    printf("Could not allocate thread.\n");
    return 1;
  }

  /* Obtain the payload as a dataspace. */

  file_t *file = client_open(argv[1], O_RDONLY);

  if (file == NULL)
  {
    printf("Could not read file: %s\n", argv[1]);
    return 1;
  }

  /* Copy the entire payload to a new dataspace. */

  char *buf;
  offset_t nread;
  offset_t region_size = find_region_size(4000000);
  l4re_ds_t region_ds;

  err = ipc_allocate_align(region_size, L4RE_RM_F_SEARCH_ADDR | L4RE_RM_F_RWX,
                           find_region_exp(4000000), (void **) &buf, &region_ds);

  if (err)
  {
    printf("Could not reserve memory.\n");
    return 1;
  }

  printf("Reading from file into %p.\n", buf);

  nread = client_read(file, buf, 4000000);

  printf("Read %ld from file.\n", nread);

  if (memcmp(buf + 0xae3, "\x31\xed", 2))
  {
    printf("Did not find expected instructions at start.\n");
    return 1;
  }

  /* UTCB location and size. */

  l4_addr_t utcb_start = Utcb_area_start;
  int utcb_log2size = find_region_exp(Default_max_threads * L4_UTCB_OFFSET);

  /* Round up to at least one page. */

  if (utcb_log2size < L4_PAGESHIFT)
    utcb_log2size = L4_PAGESHIFT;

  /* KIP allocation. */

  l4_addr_t kip_start = (l4_addr_t) l4re_kip();

  printf("KIP at %lx.\n", kip_start);

  /* Create a new task and thread. */

  l4_fpage_t utcb_fpage = l4_fpage(utcb_start, utcb_log2size, 0);

  err = l4_error(l4_factory_create_task(l4re_env()->factory, task, utcb_fpage));

  if (err)
  {
    printf("Could not create task.\n");
    return 1;
  }

  err = l4_error(l4_factory_create_thread(l4re_env()->factory, thread));

  if (err)
  {
    printf("Could not create thread.\n");
    return 1;
  }

  /* Start the pager. */

  ipc_server_config_type config;
  pthread_t pager_thread;
  pthread_attr_t attr;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  init_pager(&config, (l4_addr_t) buf, find_region_exp(file->size), 0x1000000);

  err = pthread_create(&pager_thread, &attr, ipc_server_start_mainloop, &config);

  if (err)
  {
    printf("Could not start pager thread.\n");
    return 1;
  }

  err = start_pager(&config, pager_thread);

  if (err)
  {
    printf("Could not start pager.\n");
    return 1;
  }

  /* Map the pager capability into the region manager/mapper slot. */

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(config.server, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(0x10 << L4_CAP_SHIFT, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map pager capability into task.\n");
    return 1;
  }

  /* Map the KIP into the task. */

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_fpage(kip_start, L4_PAGESHIFT, L4_FPAGE_RX),
                             kip_start));

  if (err)
  {
    printf("Could not map KIP into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(task, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_TASK_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map task capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(thread, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_THREAD_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map thread capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(l4re_env()->factory, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_FACTORY_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map factory capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(l4re_env()->log, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_LOG_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map log capability into task.\n");
    return 1;
  }

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP,
                             l4_obj_fpage(l4re_env()->scheduler, 0, L4_CAP_FPAGE_RWS),
                             l4_map_obj_control(L4_BASE_SCHEDULER_CAP, L4_MAP_ITEM_MAP)));

  if (err)
  {
    printf("Could not map scheduler capability into task.\n");
    return 1;
  }

  /* Configure the thread with the region manager acting as pager and exception
     handler. The UTCB will be situated at an address supported by a dataspace
     attached to the new task. */

  printf("Configure thread...\n");

  l4_thread_control_start();
  l4_thread_control_pager(0x10 << L4_CAP_SHIFT);
  l4_thread_control_exc_handler(0x10 << L4_CAP_SHIFT);
  l4_thread_control_bind((l4_utcb_t *) utcb_start, task);
  err = l4_error(l4_thread_control_commit(thread));

  if (err)
  {
    printf("Could not configure thread.\n");
    return 1;
  }

  /* Map the payload into the new task. */

  printf("Map %p with size %ld (2 ** %d).\n", buf, file->size, find_region_exp(file->size));

  l4_fpage_t payload_fpage = l4_fpage((l4_addr_t) buf, find_region_exp(file->size), L4_FPAGE_RX);

  err = l4_error(l4_task_map(task, L4RE_THIS_TASK_CAP, payload_fpage, 0x1000000));

  if (err)
  {
    printf("Could not map payload into task.\n");
    return 1;
  }

  /* Start the new thread. */

  printf("Schedule thread...\n");

  err = l4_error(l4_thread_ex_regs(thread, 0x1000ae3, 0x2000000 /* stack top */, 0));

  if (err)
  {
    printf("Could not set thread registers.\n");
    return 1;
  }

  printf("Run thread...\n");

  l4_sched_param_t sp = l4_sched_param(L4RE_MAIN_THREAD_PRIO);

  err = l4_error(l4_scheduler_run_thread(l4re_env()->scheduler, thread, &sp));

  if (err)
  {
    printf("Could not run thread.\n");
    return 1;
  }

  printf("Finished.\n");
  while (1);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
